@@ -45,7 +45,9 @@ struct fscrypt_context {
 	u8 flags;
 	u8 master_key_descriptor[FS_KEY_DESCRIPTOR_SIZE];
 	u8 nonce[FS_KEY_DERIVATION_NONCE_SIZE];
-} __attribute__((packed));
+};
+
+static_assert(sizeof(struct fscrypt_context) == 28, "");
 
 #define F2FS_ACL_VERSION	0x0001
 
